import monitor.*;
	
//**********************************************************************/ 
// definizione di una classe che vuole rappresentare un tipo MONITOR:   /
// deve ereditare dalla classe astratta Monitor							/
//**********************************************************************/ 

public class Frantoio extends Monitor {

    // **********************************************************************/
    // dichiarazione delle variabili di stato del MONITOR  					 /
    // **********************************************************************/

    // ----------- "COSTANTI" DEL PROBLEMA ---------------------

    int L; 				// max capacita' della cisterna in kg
    int V; 				// numero di vasche del frantoio
    int capac_vasca; 	// quantita' max di olive per vasca

    // ---------- VARIABILI DI STATO DEL MONITOR ----------------------------

    int vasche_disponibili; 		// contatore delle vasche disponibili
    int livello_cisterna; 			// quantita' di olio nella cisterna
    int contadiniSospesi;    		// contatori di processi in coda (utile per risvegliaro a gruppo
    int contadiniSospesiCisterna;	// "	"	"	"	"	"	"
 
    // **********************************************************************/
    // dichiarazione di variabili condition del MONITOR 					 /
    // **********************************************************************/

    Cond codaContadini;			// condition="presenza di un certo numero di vasche vuote"
    Cond codaPresse;			// condition="presenza di almeno una vasca piena di olive"
    Cond codaCisternaPresse;	// condition="presenza di spazio vuoto nella cisterna"
    Cond codaCisternaContadini;	// condition="presenza di olio nella cisterna"

    // **********************************************************************/
    // costruttore della classe MONITOR 									 /
    // **********************************************************************/

    public Frantoio(int CAPAC, int numVasche, int capac_cisterna) {

		// inizializzazioni delle variabili di stato
		V = numVasche;
		L = capac_cisterna;
		capac_vasca = CAPAC;
		livello_cisterna = 0;		//inizialmente la cisterna  vuota
		vasche_disponibili = V;		//inizialmente tutte le vasche sono disponibili	
		contadiniSospesi = 0;		//inizialmente non ci sono contadini che aspettano
		contadiniSospesiCisterna = 0;
		
		// creazione delle variabili condition
		codaContadini = new Cond(this);
		codaPresse = new Cond(this);
		codaCisternaPresse = new Cond(this);
		codaCisternaContadini = new Cond(this);

    }

    // Funzione utile per stampare lo stato del monitr e delle sue risorse
    private synchronized void stampaStato()
    {

		System.out.println("\n\t  VASCHE");
		for(int i=0; i<V; i++)
		    System.out.print("-----");
		System.out.println();
		for(int i=V; i>0; i--)
		    System.out.print("| "+((i>vasche_disponibili)?"1":"0")+" ");
		System.out.println("|  CISTERNA="+livello_cisterna);
		for(int i=0; i<V; i++)
		    System.out.print("-----");
		System.out.println();	
		/* 
		  VASCHE
-----------------------------------
| 0 | 0 | 0 | 0 | 0 | 0 | 0 |  CISTERNA=0
-----------------------------------
*/

    }

    // **********************************************************************/
    // definizione delle procedure e funzioni Entry del MONITOR 			 /
    // **********************************************************************/

    // Utilizzato dal CONTADINO
    public void deposita(int carico) {
    	
    	entraMonitor();

		// Per poter depositare il mio carico devo acquisire le vasche necessarie
		int vasche_richieste = carico / capac_vasca;
		// il resto della divisione deve essere zero per ipotesi

		// Il while  utile perch ogni contadino richiede un numero di vasche differenti, che non si conoscono a priori
		//  necessario quindi che sia il contadino, una volta risvegliato a riestare le condizioni necessarie
		while (true) {

			stampaStato();

		   // Se non ci sono abbastanza vasche mi devo sospendere
		    if (vasche_richieste > vasche_disponibili) {

				System.out.println(Thread.currentThread()+ ": mi sospendo perch non ci sono vasche sufficienti (me ne servono " +vasche_richieste + " ).");
				contadiniSospesi++;
				codaContadini.Wait();
				contadiniSospesi--;
				stampaStato(); //Stato prima della mia entrata nel sistema
				System.out.println(Thread.currentThread()+ ": mi risveglio per ritestare la presenza delle vasche.");
		    } 
		    else
		    	break;
		}

		System.out.println(Thread.currentThread() + ": ho trovato le vasche necessarie. Le acquisisco...");
		// Ho le risorse che cercavo. Ora le occupo.
		vasche_disponibili = vasche_disponibili - vasche_richieste;
		stampaStato(); // Stato dopo che ho occupato le risorse

		// Avendo riempito alcune vasche, devo segnalare alle presse la condizione.
		// Poiche' ogni pressa, puo' svuotare al massimo una vasca alla volta,
	    // risveglio al piu' n presse, le quali sono certe di poter acquisire ciascuna una vasca.
		// ATTENZIONE: NON E' UN ERRORE SVEGLIARE TUTTE LE PRESSE!!!
		// ma per massimizzare le risorse  meglio che ogni thread notifichi solo in base ala percentuale
		// in cui ha cambiato lui lo stato del monitor 
		
		System.out.println(Thread.currentThread() + ": risveglio al pi "+ vasche_richieste + "presse in attesa...");
		for (int i = 0; i < vasche_richieste && codaPresse.queue(); i++)
		    codaPresse.Signal();

		esciMonitor();
		
    }

    // Utilizzata dal CONTADINO
    public void ritira(int olio) {

    	entraMonitor();

		stampaStato();
		
		// Se non  ancora pronta la quantit di olio che aspetto, mi sospendo
		// Anche qui  utile un while perch solo il contadino conosce il suo livello di olio richiesto
		// ed  giusto che sia lui a ricontrollare le condizioni, una volta che viene risvegliato
		while (livello_cisterna < olio) {	    
		    System.out.println(Thread.currentThread() + ": non  ancora pronto il mio olio. Mi sospendo...");
		    contadiniSospesiCisterna++;
		    codaCisternaContadini.Wait();
		    contadiniSospesiCisterna--;
		    stampaStato();
		    System.out.println(Thread.currentThread() + ": mi risveglio per ritestare il livello della cisterna...");
		}

		System.out.println(Thread.currentThread()+ ": ritiro il mio olio dalla cisterna...");
		// libero le risorse
		livello_cisterna -= olio;
	
		stampaStato();

		// SONO LE PRESSE CHE UTILIZZANO LA CISTERNA PER VERSARE OLIO quindi devo risvegliare loro 

		// Posso sapere in anticipo quante presse devo al piu' risvegliare.
		int presseMax = olio / capac_vasca; // perche'ogni pressa riversa nella cisterna sempre
	                                        // capac_vasca kg di olio
		System.out.println(Thread.currentThread() + ": risveglio al pi "+ presseMax + " presse che erano in attesa della cisterna...");
		for (; presseMax > 0; presseMax--)
		    codaCisternaPresse.Signal();

		esciMonitor();

    }

    // Utilizzato dalla PRESSA
    public void caricaUnaVasca() {

    	entraMonitor();

		stampaStato();

		// CHI MI RISVEGLIA HA LA CERTEZZA DELLA MIA ACQUISIZIONE
		
		// MI sospendo se non ci sono vasche da scaricare
		if (vasche_disponibili == V) {
		    System.out.println(Thread.currentThread() + ": non ci sono vasche da scaricare. Mi addormento...");
		    codaPresse.Wait();
		}

		// Ho trovato una vasca non vuota. La svuoto cosi' posso lavorarla.
		System.out.println(Thread.currentThread() + ": svuoto una vasca...");
		vasche_disponibili++;

		stampaStato();

		// IL CONTADINO (nella sua fase di produttore) E' L'UNICO INTERESSATO
		// ALLA MODIFICA DELLO STATO DELLA RISORSA "VASCHE" perch deve riempirle
	
		// Ho liberato una vasca, quindi posso risvegliare eventuali contadini
	    // per lasciarli ritestare le condizioni. Uso un risveglio a gruppo per
	    // lasciare che ciascun contadino ritesti le sue condizioni. 
		// --> STARVATION DEI CONTADINI CON MOLTE OLIVE DA CARICARE!!

		int sospesi = contadiniSospesi;
		for (; sospesi > 0; sospesi--)
		    codaContadini.Signal();

		esciMonitor();

    }

    // Utilizzato dalla PRESSA
    public void riempiCisterna() {

    	entraMonitor();

		// Dobbiamo verificare di avere spazio nella cisterna
		// Non c'e' bisogno di while, perche' e' nota e FISSA la quantita' di olio
	    // da riversare nella cisterna

		stampaStato();

		// Mi sospendo se non c' spazio nella cisterna per l'olio che produco
		if (livello_cisterna + capac_vasca > L) {
		    System.out.println(Thread.currentThread() + ": non c' spazio sufficiente nella cisterna. Mi sospendo...");
		    codaCisternaPresse.Wait();
		}

		System.out.println(Thread.currentThread() + ": deposito l'olio che ho prodotto nella cisterna");
		livello_cisterna += capac_vasca;

		stampaStato();

		// risveglio a gruppo perche' non so ogni contadino quanto olio prelevera'
	    // dalla cisterna (i contadini ritestano la condizione di entrata)
		
		// NB non decremento qui contadiniSospesiCisterna, ma viene fatto dai contadini stessi
		int sospesi = contadiniSospesiCisterna;
		for (; sospesi > 0; sospesi--)
		    codaCisternaContadini.Signal();

		esciMonitor();

    }

} 

